Imports System.Runtime.InteropServices

Public Class FormMain
  Private Stream As IO.FileStream

  Private Sub FormMain_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
    Dim UODemoPath As String = Nothing

    ' Get the last opened path
    Try
      UODemoPath = GetSetting("UO Demo Archiver", "DAT", "Path").Trim
    Catch
      UODemoPath = ""
    End Try

    If UODemoPath <> "" Then
      ' Accept any stored path without checking it
      Me.TextBox2.Text = UODemoPath
    Else
      Try
        ' Get the path to the Client EXE
        ' And modify it to point to the UODEMO.DAT file
        UODemoPath = My.Computer.Registry.GetValue("HKEY_LOCAL_MACHINE\SOFTWARE\Origin Worlds Online\Ultima Online\1.0", "ExePath", Nothing)
        UODemoPath = IO.Path.GetDirectoryName(UODemoPath)
        UODemoPath = IO.Path.Combine(UODemoPath, "UoDemo\UODEMO.DAT")
        If IO.File.Exists(UODemoPath) Then
          ' Only accept this path if the UODEMO.DAT file exists
          Me.TextBox2.Text = UODemoPath
        End If
      Catch
      End Try
    End If

    ' By default, we set for delete
    Me.Button_ToggleDelete.Text = "Delete"
  End Sub

  Private Sub Button_Open_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button_Open.Click
    ' INTERFACE: disable all elements
    Me.TextBox2.Enabled = False
    Me.Button_SelectFile.Enabled = False
    Me.Button_Open.Enabled = False
    Me.Button_Close.Enabled = False
    Me.Button_QuickView.Enabled = False
    Me.Button_Extract.Enabled = False
    Me.Button_ExtractAll.Enabled = False
    Me.Button_Replace.Enabled = False
    Me.Button_ToggleDelete.Enabled = False

    ' INTERFACE: clear the existing list with items
    Me.ListView1.SelectedItems.Clear()
    Me.ListView1.Items.Clear()

    ' Store the last opened path (even on failure)
    Dim UODemoPath As String = Me.TextBox2.Text.Trim
    Try
      SaveSetting("UO Demo Archiver", "DAT", "Path", UODemoPath)
    Catch
    End Try

    Try
      Stream = New IO.FileStream(UODemoPath, IO.FileMode.Open, IO.FileAccess.ReadWrite, IO.FileShare.Read, UODEMO_FILEHEADER_SIZE)
    Catch ex As Exception
      ' An error occurred while opening the file
      Stream = Nothing

      ' INTERFACE: display the error message
      MsgBox(ex.Message)

      ' INTERFACE: we can open files again
      Me.TextBox2.Enabled = True
      Me.Button_SelectFile.Enabled = True
      Me.Button_Open.Enabled = True
      Exit Sub
    End Try
    
    Try
      Do
        ' Read each header
        Dim Header As UODEMO_FILEHEADER = ReadHeader(Stream)
        If Header.FilePath = "@@@.@@@" Then
          ' End reached
          Exit Do
        End If

        ' Display the header
        Dim LV As ListViewItem = Me.ListView1.Items.Add(Header.FilePath)
        LV.SubItems.Add("0x" & Header.FilePointer.ToString("X"))
        LV.SubItems.Add("0x" & Header.ReservedSize.ToString("X"))
        LV.SubItems.Add(Header.IsUnknown2.ToString("X"))
        LV.SubItems.Add("0x" & Header.FileSize.ToString("X"))
        LV.SubItems.Add("0x" & Header.FilePosition.ToString("X"))
        LV.Tag = Header
      Loop

      ' NOTE:
      ' The file pointer now points to the beginning of the file contents

      ' INTERFACE: we cannot open files we can close it and work on it :)
      Me.Button_Close.Enabled = True
      Me.Button_ExtractAll.Enabled = True
    Catch ex As Exception
      ' An error occurred while reading the file
      Try
        Stream.Close()
      Catch
      End Try
      Stream = Nothing

      ' INTERFACE: display the error message
      MsgBox(ex.Message)

      ' INTERFACE: we can open files again
      Me.TextBox2.Enabled = True
      Me.Button_SelectFile.Enabled = True
      Me.Button_Open.Enabled = True
      Exit Sub
    End Try
  End Sub

  Private Sub Button_Close_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button_Close.Click
    ' INTERFACE: disable all elements
    Me.TextBox2.Enabled = False
    Me.Button_SelectFile.Enabled = False
    Me.Button_Open.Enabled = False
    Me.Button_Close.Enabled = False
    Me.Button_QuickView.Enabled = False
    Me.Button_Extract.Enabled = False
    Me.Button_ExtractAll.Enabled = False
    Me.Button_Replace.Enabled = False
    Me.Button_ToggleDelete.Enabled = False

    ' INTERFACE: clear the existing list with items
    Me.ListView1.SelectedItems.Clear()
    Me.ListView1.Items.Clear()

    Try
      ' Close the file, ignore any errors
      Stream.Close()
    Catch
    End Try
    Stream = Nothing

    ' INTERFACE: we can open files again
    Me.TextBox2.Enabled = True
    Me.Button_SelectFile.Enabled = True
    Me.Button_Open.Enabled = True
  End Sub

  Private Sub Button_SelectFile_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button_SelectFile.Click
    Try
      ' Use a Open File Dialog to get a new file name
      Dim GetFile As New OpenFileDialog
      GetFile.CheckFileExists = True
      GetFile.DefaultExt = "DAT"
      GetFile.Filter = "DAT files (*.DAT)|*.DAT|All Files (*.*)|*.*"
      GetFile.FilterIndex = 0
      GetFile.FileName = Me.TextBox2.Text.Trim
      If GetFile.ShowDialog() = Windows.Forms.DialogResult.OK Then
        Me.TextBox2.Text = GetFile.FileName
      End If
    Catch ex As Exception
      ' Error!
      MsgBox(ex.Message)
    End Try
  End Sub

  Private Sub ListView1_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ListView1.SelectedIndexChanged
    Me.Button_QuickView.Enabled = Me.ListView1.SelectedItems.Count = 1
    Me.Button_Extract.Enabled = Me.ListView1.SelectedItems.Count = 1
    Me.Button_Replace.Enabled = Me.ListView1.SelectedItems.Count = 1
    Me.Button_ToggleDelete.Enabled = Me.ListView1.SelectedItems.Count = 1

    If Me.ListView1.SelectedItems.Count = 1 Then
      Dim Header As UODEMO_FILEHEADER
      Header = Me.ListView1.SelectedItems(0).Tag

      If Header.FilePath.Length > 0 Then
        If Header.FilePath(0) = "|"c Then
          Me.Button_ToggleDelete.Text = "Undelete"
        Else
          Me.Button_ToggleDelete.Text = "Delete"
        End If
      Else
        Me.Button_ToggleDelete.Text = "Delete"

        ' NOTE: filepath seems to be invalid, don't allow deletion
        Me.Button_ToggleDelete.Enabled = False
      End If
    Else
      Me.Button_ToggleDelete.Text = "Delete"
    End If
  End Sub

  Private Sub Button_QuickView_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button_QuickView.Click
    Dim TempPath As String
    Dim Header As UODEMO_FILEHEADER
    Try
      TempPath = My.Computer.FileSystem.GetTempFileName
      Header = Me.ListView1.SelectedItems(0).Tag
    Catch
      Exit Sub
    End Try

    Try
      ' Read the file contents
      Dim FileContent() As Byte = UODAT.ReadFile(Stream, Header)

      ' Write the decrypted data to a temponary file)
      Dim FileWriter As New IO.FileStream(TempPath, IO.FileMode.Create, IO.FileAccess.Write, IO.FileShare.Read)
      Dim BinaryWriter As New IO.BinaryWriter(FileWriter)
      If FileContent IsNot Nothing Then
        BinaryWriter.Write(FileContent, 0, FileContent.Length)
      End If
      FileWriter.Close()

      ' Display using notepad
      Shell("notepad.exe """ & TempPath & """", AppWinStyle.NormalFocus)
    Catch ex As Exception
      MsgBox(ex.Message)
    End Try
  End Sub

  Private Sub Button_Extract_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button_Extract.Click
    Dim Header As UODEMO_FILEHEADER
    Try
      Header = Me.ListView1.SelectedItems(0).Tag
    Catch
      Exit Sub
    End Try

    Try
      ' Remove .Q
      Dim FilePath As String = Header.FilePath
      If Microsoft.VisualBasic.Right(FilePath, 2).ToLowerInvariant = ".q" Then
        FilePath = FilePath.Substring(0, FilePath.Length - 2)
      End If

      ' Get the extension (keep casing)
      Dim FileExtension As String = IO.Path.GetExtension(FilePath)
      If Microsoft.VisualBasic.Left(FileExtension, 1) = "." Then
        FileExtension = FileExtension.Substring(1, FileExtension.Length - 1)
      End If

      ' Get the target filename
      Dim GetFile As New SaveFileDialog
      GetFile.CheckPathExists = True
      GetFile.DefaultExt = FileExtension
      GetFile.Filter = "*." & FileExtension & " Files (*." & FileExtension & ")|*." & FileExtension & "|All Files (*.*)|*.*"
      GetFile.FilterIndex = 0
      GetFile.FileName = IO.Path.GetFileName(FilePath)
      GetFile.OverwritePrompt = True
      If GetFile.ShowDialog() <> Windows.Forms.DialogResult.OK Then
        Exit Sub
      End If
      FilePath = GetFile.FileName

      ' Read the file contents
      Dim FileContent() As Byte = UODAT.ReadFile(Stream, Header)

      ' Write the decrypted data to a temponary file)
      Dim FileWriter As New IO.FileStream(FilePath, IO.FileMode.Create, IO.FileAccess.Write, IO.FileShare.Read)
      Dim BinaryWriter As New IO.BinaryWriter(FileWriter)
      If FileContent IsNot Nothing Then
        BinaryWriter.Write(FileContent, 0, FileContent.Length)
      End If
      FileWriter.Close()
    Catch ex As Exception
      MsgBox(ex.Message)
    End Try
  End Sub

  Private Sub Button_Replace_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button_Replace.Click
    Dim Header As UODEMO_FILEHEADER
    Try
      ' Get the header
      Dim lv As ListViewItem = Me.ListView1.SelectedItems(0)
      Header = lv.Tag

      ' Enable seperate styling for subitems
      lv.UseItemStyleForSubItems = False
    Catch
      Exit Sub
    End Try

    Dim NewColor As Color = Color.Black
    Try
      ' Remove .Q
      Dim FilePath As String = Header.FilePath
      If Microsoft.VisualBasic.Right(FilePath, 2).ToLowerInvariant = ".q" Then
        FilePath = FilePath.Substring(0, FilePath.Length - 2)
      End If

      ' Get the extension (keep casing)
      Dim FileExtension As String = IO.Path.GetExtension(FilePath)
      If Microsoft.VisualBasic.Left(FileExtension, 1) = "." Then
        FileExtension = FileExtension.Substring(1, FileExtension.Length - 1)
      End If

      ' Use a Open File Dialog to get the file name of the new file
      Dim GetFile As New OpenFileDialog
      GetFile.CheckFileExists = True
      GetFile.DefaultExt = Nothing
      GetFile.Filter = "*." & FileExtension & ".Q Files (*." & FileExtension & ".Q)|*." & FileExtension & ";*." & FileExtension & ".Q|All Files (*.*)|*.*"
      GetFile.FilterIndex = 0
      GetFile.FileName = Nothing
      If GetFile.ShowDialog() <> Windows.Forms.DialogResult.OK Then
        Exit Sub
      End If

      ' Write
      Call UODAT.WriteFile(Stream, Header, GetFile.FileName)

      ' Tell the user everything went ok
      MsgBox("Success!" & vbNewLine & vbNewLine & "The file """ & Header.FilePath & """ has been replaced with the contents of """ & GetFile.FileName & """.")
    Catch ex As Exception
      ' Error!
      MsgBox("Error!" & vbNewLine & vbNewLine & ex.Message)

      ' Change the color to red to indicate an error
      NewColor = Color.Red
    End Try

    ' Update the list view
    Try
      Dim tmps As String
      Dim lv As ListViewItem = Me.ListView1.SelectedItems(0)

      ' Make the modified file name also bold
      lv.SubItems(0).ForeColor = NewColor
      lv.SubItems(0).Font = New Font(lv.SubItems(0).Font, FontStyle.Bold)

      ' Set color to orange if the value changed (FilePointer)
      tmps = "0x" & Header.FilePointer.ToString("X")
      If lv.SubItems(1).Text <> tmps Then
        lv.SubItems(1).Text = tmps
        lv.SubItems(1).ForeColor = Color.Orange
      Else
        lv.SubItems(1).ForeColor = Color.Black
      End If

      ' Set color to orange if the value changed (ReservedSize)
      tmps = "0x" & Header.ReservedSize.ToString("X")
      If lv.SubItems(2).Text <> tmps Then
        lv.SubItems(2).Text = tmps
        lv.SubItems(2).ForeColor = Color.Orange
      Else
        lv.SubItems(2).ForeColor = Color.Black
      End If

      ' Set color to orange if the value changed (FileSize)
      tmps = "0x" & Header.FileSize.ToString("X")
      If lv.SubItems(4).Text <> tmps Then
        lv.SubItems(4).Text = tmps
        lv.SubItems(4).ForeColor = Color.DarkOrange
      Else
        lv.SubItems(4).ForeColor = Color.Black
      End If
    Catch
    End Try
  End Sub

  Private Sub Button_ToggleDelete_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button_ToggleDelete.Click
    Dim lv As ListViewItem = Nothing
    Dim Header As UODEMO_FILEHEADER
    Try
      lv = Me.ListView1.SelectedItems(0)
      Header = lv.Tag
    Catch
      Exit Sub
    End Try

    ' Abort if we have an invalid path
    If Header.FilePath.Length = 0 Then
      Exit Sub
    End If

    ' If the first character is "." we make it "|" and otherwise around
    If False Then
    ElseIf Header.FilePath(0) = "|"c Then
      Mid(Header.FilePath, 1, 1) = "."c
      Me.Button_ToggleDelete.Text = "Delete"
    ElseIf Header.FilePath(0) = "."c Then
      Mid(Header.FilePath, 1, 1) = "|"c
      Me.Button_ToggleDelete.Text = "Undelete"
    Else
      ' Unknown situation
      MsgBox("The first character in the path is not as expected!", MsgBoxStyle.Information Or MsgBoxStyle.OkOnly)
      Exit Sub
    End If

    Try
      ' Write the new header to disk
      Call WriteHeader(Stream, Header)
    Catch ex As Exception
      ' Error!
      MsgBox("Error!" & vbNewLine & vbNewLine & ex.Message)
    End Try

    ' Update the listview
    lv.UseItemStyleForSubItems = False
    lv.Text = Header.FilePath
    lv.ForeColor = Color.Orange
  End Sub

  Private Sub Button_CreateEmpty_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button_CreateEmpty.Click
    ' We cannot operate on a file that's already open
    If Stream IsNot Nothing Then
      MsgBox("You must close the file first!")
      Exit Sub
    End If

    ' Get the path but do not store
    Dim UODemoPath As String = Me.TextBox2.Text.Trim

    Try
      ' Create the file, overwriting any existing file without pardon
      Stream = New IO.FileStream(UODemoPath, IO.FileMode.Create, IO.FileAccess.Write, IO.FileShare.Read, UODEMO_FILEHEADER_SIZE)

      ' Ask to write a header
      If MsgBox("The file has been created!" & vbNewLine & "Do you want to add an @@@.@@@ marker?", MsgBoxStyle.Question Or MsgBoxStyle.DefaultButton1 Or MsgBoxStyle.YesNo) = MsgBoxResult.Yes Then
        Dim emptyheader As New UODEMO_FILEHEADER
        emptyheader.FilePath = "@@@.@@@"
        WriteHeader(Stream, emptyheader)
      End If
    Catch ex As Exception
      MsgBox(ex.Message)
    End Try

    ' Close the file, ignore any errors
    Try
      Stream.Close()
    Catch
    End Try

    ' Stream must be Nothing at exit
    Stream = Nothing
  End Sub

  Private Sub Button_ExtractAll_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button_ExtractAll.Click
    ' Get the target directory for extraction
    Dim x As New FolderBrowserDialog
    x.Description = "Select the target directory"
    x.ShowNewFolderButton = True
    If x.ShowDialog(Me) = Windows.Forms.DialogResult.OK Then
      ' 
      Dim allok As Boolean = True

      ' Handle each item in the list (selected or not)
      Dim Header As UODEMO_FILEHEADER, TargetFilePath As String
      For Each lv As ListViewItem In Me.ListView1.Items
        Try
          Header = lv.Tag
        Catch
          Exit For
        End Try

        ' Build the target file path
        TargetFilePath = x.SelectedPath

        ' Ignore deleted files!
        If Header.FilePath(0) <> "|"c Then
          ' Get and create the sub directory if any
          Dim SubDirs() As String = Header.FilePath.Split("/"c)
          For i As Integer = SubDirs.GetLowerBound(0) To SubDirs.GetUpperBound(0)
            Dim SubDir As String = SubDirs(i)
            If SubDir = "." Then
              ' Ignore
            ElseIf SubDir = ".." Then
              ' Ignore only in the beginning
              If i > 0 Then
                TargetFilePath = IO.Path.Combine(TargetFilePath, SubDir)
              End If
            ElseIf SubDir.ToLowerInvariant = ".rundir" Then
              ' Ignore
            Else
              TargetFilePath = IO.Path.Combine(TargetFilePath, SubDir)
              If i < SubDirs.GetUpperBound(0) Then
                If Not IO.Directory.Exists(TargetFilePath) Then
                  ' This is a subdirectory
                  Try
                    IO.Directory.CreateDirectory(TargetFilePath)
                  Catch ex As Exception
                    MsgBox("Unable to create required subdirectory: " & TargetFilePath & vbNewLine & "Reason:" & ex.Message, MsgBoxStyle.Critical Or MsgBoxStyle.OkOnly)
                    allok = False
                    Exit For
                  End Try
                End If
              End If
            End If
          Next

          ' Read the file contents
          Dim FileContent() As Byte = UODAT.ReadFile(Stream, Header)

          Try
            ' Write the decrypted data to a temponary file)
            Dim FileWriter As New IO.FileStream(TargetFilePath, IO.FileMode.Create, IO.FileAccess.Write, IO.FileShare.Read)
            Dim BinaryWriter As New IO.BinaryWriter(FileWriter)
            If FileContent IsNot Nothing Then
              BinaryWriter.Write(FileContent, 0, FileContent.Length)
            End If
            FileWriter.Close()
          Catch ex As Exception
            MsgBox("Unable to create file: " & TargetFilePath & vbNewLine & "Reason:" & ex.Message, MsgBoxStyle.Critical Or MsgBoxStyle.OkOnly)
            allok = False
            Exit For
          End Try
        End If
      Next

      ' Give a nice message if we are finished
      If allok Then
        MsgBox("Done!", MsgBoxStyle.Information Or MsgBoxStyle.OkOnly)
      End If
    End If
  End Sub
End Class