Module UODAT
  Public Const MAXPATH As Integer = 260
  Public Const UODEMO_FILEHEADER_SIZE As Integer = 260 + 5 * 4 ' &H118

  Private Class UODEMO_FILEHEADER_ENCRYPTED
    Public DWORDs(0 To UODEMO_FILEHEADER_SIZE / 4 - 1) As UInt32
  End Class

  ' NOTE: this must be a Class, so changes made will be made everywhere
  ' if this were a Structure then we would have to keep a Header database or something
  ' http://msdn.microsoft.com/en-us/library/aa289521(VS.71).aspx
  Public Class UODEMO_FILEHEADER
    ' Fields used by us
    Public HeaderPointer As Long

    ' "Official" fields
    Public FilePath As String
    Public FilePointer As UInt32
    Public ReservedSize As UInt32 ' seems to mean MaximumSize
    Public IsUnknown2 As UInt32 ' meaning seems to be IsReadOnly
    Public FileSize As UInt32
    Public FilePosition As UInt32 ' meaning is file position, used by the uodemo internally, not important here!
  End Class

  Public Function ReadHeader(ByRef Stream As IO.Stream) As UODEMO_FILEHEADER
    Dim HeaderE As New UODEMO_FILEHEADER_ENCRYPTED
    Dim Header As New UODEMO_FILEHEADER

    ' Store the current file pointer
    Dim HeaderPointer As Integer = Stream.Position

    ' Open the file and read the encrypted bytes
    Dim tempbuffer(0 To UODEMO_FILEHEADER_SIZE - 1) As Byte
    Dim BinaryReader As New IO.BinaryReader(Stream)
    BinaryReader.Read(tempbuffer, 0, tempbuffer.Length)

    ' Copy these bytes to a decryption buffer
    ' NOTE: if VB.NET would support arrays in unions I would implement this differently!
    Buffer.BlockCopy(tempbuffer, 0, HeaderE.DWORDs, 0, tempbuffer.Length)
    GOST_UODAT.UODAT_gostofb(HeaderE.DWORDs, HeaderE.DWORDs, HeaderE.DWORDs.Length / 2)

    ' Copy the decrypted bytes (of the filepath) back to a temponary buffer
    Buffer.BlockCopy(HeaderE.DWORDs, 0, tempbuffer, 0, MAXPATH)

    ' Convert the decrypted bytes to a string
    Dim tmpbuilder As New System.Text.StringBuilder(MAXPATH)
    For Each b As Byte In tempbuffer
      If b = 0 Then
        Exit For
      End If
      tmpbuilder = tmpbuilder.Append(Chr(b))
    Next

    ' And setup the UODEMO_HEADER structure
    Header.HeaderPointer = HeaderPointer
    Header.FilePath = tmpbuilder.ToString
    Header.FilePointer = HeaderE.DWORDs(65)
    Header.ReservedSize = HeaderE.DWORDs(66)
    Header.IsUnknown2 = HeaderE.DWORDs(67)
    Header.FileSize = HeaderE.DWORDs(68)
    Header.FilePosition = HeaderE.DWORDs(69)

    Return Header
  End Function

  Public Sub WriteHeader(ByRef Stream As IO.Stream, ByRef Header As UODEMO_FILEHEADER)
    ' Write the data to the file
    ' IMPORTANT: maintain the current position even on error!
    Dim ex As Exception = Nothing

    Try
      ' Remember the file position
      Dim CurrentPosition As Long = Stream.Position
      Stream.Seek(Header.FilePointer, IO.SeekOrigin.Current)

      Try
        Dim tempbuffer(0 To UODEMO_FILEHEADER_SIZE - 1) As Byte
        Dim HeaderE As New UODEMO_FILEHEADER_ENCRYPTED

        ' This is a bit rough but it does the job
        Dim index As Integer = 0
        For Each c As Char In Header.FilePath
          tempbuffer(index) = Asc(c)
          index += 1
        Next

        ' Ensure 0-terminator!
        tempbuffer(MAXPATH - 1) = 0

        ' STEP 1: Create a new header with our data
        Buffer.BlockCopy(tempbuffer, 0, HeaderE.DWORDs, 0, MAXPATH)
        HeaderE.DWORDs(65) = Header.FilePointer
        HeaderE.DWORDs(66) = Header.ReservedSize
        HeaderE.DWORDs(67) = Header.IsUnknown2
        HeaderE.DWORDs(68) = Header.FileSize
        HeaderE.DWORDs(69) = Header.FilePosition

        ' STEP 2: Encrypt
        GOST_UODAT.UODAT_gostofb(HeaderE.DWORDs, HeaderE.DWORDs, HeaderE.DWORDs.Length / 2)

        ' STEP 3: Copy the decrypted bytes back to a temponary buffer
        Buffer.BlockCopy(HeaderE.DWORDs, 0, tempbuffer, 0, tempbuffer.Length)

        ' STEP 4: Write
        Stream.Seek(Header.HeaderPointer, IO.SeekOrigin.Begin)
        Stream.Write(tempbuffer, 0, tempbuffer.Length)
      Catch tmpex As Exception
        ex = tmpex
      End Try

      ' Restore file position
      Stream.Seek(CurrentPosition, IO.SeekOrigin.Begin)
    Catch
    End Try

    ' Forward the error if any
    If ex IsNot Nothing Then
      Throw ex
    End If
  End Sub

  Public Function ReadFile(ByRef Stream As IO.Stream, ByRef Header As UODEMO_FILEHEADER) As Byte()
    ' Verify the file size
    Dim FileSize As Long = Header.FileSize
    If FileSize = 0 Then
      Return Nothing
    End If

    ' Calculate the buffer size, this must be 8 byte aligned
    ' NOTE: the demo used a fixed buffer size of 4KB
    Dim BufferSize As Long = Header.FileSize
    If BufferSize Mod 8 > 0 Then
      BufferSize += 8 - (BufferSize Mod 8)
    End If

    ' Temponary buffers
    Dim Data(BufferSize - 1) As Byte
    Dim DWORDs(&H1000 / 4 - 1) As UInt32

    ' Read the data from the file
    ' IMPORTANT: maintain the current position even on error!
    Dim ex As Exception = Nothing
    Try
      Dim CurrentPosition As Long = Stream.Position
      Stream.Seek(Header.FilePointer, IO.SeekOrigin.Current)
      Try
        Stream.Read(Data, 0, FileSize)
      Catch extmp As Exception
        ex = extmp
      End Try
      Stream.Seek(CurrentPosition, IO.SeekOrigin.Begin)
    Catch
    End Try
    If ex IsNot Nothing Then
      ' Forward error
      Throw ex
    End If

    ' Decrypt in blocks of 4KB (which seems to be the size that the UO Demo caches its data with)
    For Index As Integer = 0 To FileSize - 1 Step 0
      ' Calculate the number of bytes to handle
      Dim Remainder As Integer = &H1000
      If Remainder > FileSize - Index Then
        Remainder = FileSize - Index

        ' Fix-Up so UODAT_gostofb can do its work correctly...
        If Remainder Mod 8 > 0 Then
          Remainder += 8 - Remainder Mod 8
        End If
      End If

      ' Decrypt (using DWORDs buffer)
      ' (in C we would simply cast to another type, in .NET this is not possible)
      Buffer.BlockCopy(Data, Index, DWORDs, 0, Remainder)
      GOST_UODAT.UODAT_gostofb(DWORDs, DWORDs, Remainder \ 8)
      Buffer.BlockCopy(DWORDs, 0, Data, Index, Remainder)

      ' Next
      Index = Index + Remainder
    Next

    ' Extract the actual length
    ' (it is stored in the last decrypted dword)
    Dim ActualLength As Long
    If FileSize Mod &H1000 = 0 Then
      ActualLength = DWORDs(&H1000 - 1)
    ElseIf FileSize Mod 4 = 0 Then
      ActualLength = DWORDs((FileSize Mod &H1000) \ 4 - 1)
    Else
      ' NOTE: this situation shouldn't occur but be support it anyways
      Buffer.BlockCopy(Data, FileSize - 4, DWORDs, 0, 4)
      ActualLength = DWORDs(0)
    End If

    ' Verify the actual length
    If ActualLength > FileSize - 4 Then
      Throw New Exception("Decryption Error! (" & ActualLength & " > " & FileSize - 4)
    End If
    If ActualLength = 0 Then
      Return Nothing
    End If

    ' Return the final array
    ReDim Preserve Data(ActualLength - 1)
    Return Data
  End Function

  Public Sub WriteFile(ByRef Stream As IO.Stream, ByRef Header As UODEMO_FILEHEADER, ByVal Data() As Byte)
    ' If Data is Nothing then we make the file a zero length file
    If Data Is Nothing Then
      ' Set the reserved size if it hasn't been done yet
      If Header.ReservedSize = 0 Then
        Header.ReservedSize = Header.FileSize
      End If

      ' Set the file size to zero
      Header.FileSize = 0
      Exit Sub
    End If

    ' Add at least 4 bytes to the array (for storing the actual length)
    Dim ActualLength As Integer = Data.Length
    Dim TotalLength As Integer = ActualLength + 4

    ' Align the file length to 4 bytes (1 DWORD)
    If TotalLength Mod 4 > 0 Then
      TotalLength = TotalLength + 4 - (TotalLength Mod 4)
    End If

    ' Align the internal buffer size to 8 bytes (2 DWORD's)
    Dim BufferSize As Long = TotalLength
    If BufferSize Mod 8 > 0 Then
      BufferSize = BufferSize + 8 - (BufferSize Mod 8)
    End If

    ' Modify the buffer
    ReDim Preserve Data(BufferSize - 1)

    ' Store the length in the buffer
    Data(TotalLength - 1) = (ActualLength >> 24) And &HFF
    Data(TotalLength - 2) = (ActualLength >> 16) And &HFF
    Data(TotalLength - 3) = (ActualLength >> 8) And &HFF
    Data(TotalLength - 4) = (ActualLength >> 0) And &HFF

    ' Temponary buffer
    Dim DWORDs(&H1000 / 4 - 1) As UInt32

    ' Encrypt in blocks of 4KB (which seems to be the size that the UO Demo caches its data with)
    For Index As Integer = 0 To TotalLength - 1 Step 0
      ' Calculate the number of bytes to handle
      Dim Remainder As Integer = &H1000
      If Remainder > TotalLength - Index Then
        Remainder = TotalLength - Index

        ' Fix-Up so UODAT_gostofb can do its work correctly...
        If Remainder Mod 8 > 0 Then
          Remainder += 8 - Remainder Mod 8
        End If
      End If

      ' Decrypt (using DWORDs buffer)
      ' (in C we would simply cast to another type, in .NET this is not possible)
      Buffer.BlockCopy(Data, Index, DWORDs, 0, Remainder)
      GOST_UODAT.UODAT_gostofb(DWORDs, DWORDs, Remainder \ 8)
      Buffer.BlockCopy(DWORDs, 0, Data, Index, Remainder)

      ' Next
      Index = Index + Remainder
    Next

    ' Will the file fit?
    Dim WillFitAtCurrentLocation As Boolean = False
    If Header.ReservedSize = 0 Then
      If TotalLength <= Header.FileSize Then
        WillFitAtCurrentLocation = True
      End If
    Else
      If TotalLength <= Header.ReservedSize Then
        WillFitAtCurrentLocation = True
      End If
    End If

    If Not WillFitAtCurrentLocation Then
      ' The file doesn't fit in the current block thus...
      ' ...we will move the file to the stream and assume a maximum growing size of 30%
      Header.FilePointer = Stream.Length
      Header.ReservedSize = TotalLength * 1.3
      Header.ReservedSize = Header.ReservedSize - Header.ReservedSize Mod 8

      ' We modified the file pointer so now the file with fit :-)
      WillFitAtCurrentLocation = True
    End If

    If WillFitAtCurrentLocation Then
      ' Write the data to the file
      ' IMPORTANT: maintain the current position even on error!
      Dim ex As Exception = Nothing
      Try
        Dim CurrentPosition As Long = Stream.Position
        Stream.Seek(Header.FilePointer, IO.SeekOrigin.Current)
        Try
          ' Write the data
          Stream.Write(Data, 0, TotalLength)

          ' Write the reserved data
          If Header.ReservedSize > 0 AndAlso Header.ReservedSize > TotalLength Then
            Dim ReservedBytesToWrite As Long = Header.ReservedSize - TotalLength
            Do While ReservedBytesToWrite > 0
              Dim CurrentBytesToWrite As Long = ReservedBytesToWrite
              If CurrentBytesToWrite > Data.Length Then
                CurrentBytesToWrite = Data.Length
              End If
              Stream.Write(Data, 0, CurrentBytesToWrite)
              ReservedBytesToWrite -= CurrentBytesToWrite
            Loop
          End If

          ' ON SUCCESS: If the file size changes then we must set the reserved size
          If (Header.ReservedSize = 0) AndAlso (Header.FileSize <> TotalLength) Then
            Header.ReservedSize = Header.FileSize
          End If

          ' ON SUCCESS: Indicate the new file size
          Header.FileSize = TotalLength

          ' Go modify the header
          Call WriteHeader(Stream, Header)
        Catch extmp As Exception
          ex = extmp
        End Try
        Stream.Seek(CurrentPosition, IO.SeekOrigin.Begin)
      Catch
      End Try
      If ex IsNot Nothing Then
        ' Forward error
        Throw ex
      End If
    Else
      Throw New Exception("Sorry, an unexpected situation occurred which is preventing us from writing the file!")
    End If
  End Sub

  Public Sub WriteFile(ByRef Stream As IO.Stream, ByRef Header As UODEMO_FILEHEADER, ByVal NewPath As String)
    ' Open the file
    Dim NewStream As IO.Stream = New IO.FileStream(NewPath, IO.FileMode.Open, IO.FileAccess.Read, IO.FileShare.ReadWrite)
    Try
      If NewStream.Length = 0 Then
        ' File is empty
        Call WriteFile(Stream, Header, CType(Nothing, Array))
      Else
        ' Read the file into memory
        Dim BinaryReader As New IO.BinaryReader(NewStream)
        Dim Data() As Byte = BinaryReader.ReadBytes(NewStream.Length)
        Call WriteFile(Stream, Header, Data)
      End If
    Catch ex As Exception
      ' Close the file even on error
      Try
        NewStream.Close()
      Catch
      End Try

      ' Forward the error
      Throw ex
    End Try
  End Sub
End Module
